local ADDONNAME, ns = ...

-- create addon
local Addon = LibStub("AceAddon-3.0"):NewAddon(ADDONNAME, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")

-- load locale
local L = LibStub("AceLocale-3.0"):GetLocale(ADDONNAME)

-- load libs
local AC = LibStub("AceConfig-3.0")
local ACD = LibStub("AceConfigDialog-3.0")

-- WoW APIs
local _G = _G

-- Lua API
local select, type, pairs, next, wipe = select, type, pairs, next, wipe
local tostring, tonumber = tostring, tonumber

-- GLOBALS: LibStub, WhisperInviteSettings

function Addon:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("WhisperInviteDB", ns.defaults, true)
    self.db.RegisterCallback(self, "OnProfileCopied", "OnDBUpdate")
    --self.db.RegisterCallback(self, "OnDatabaseReset", "OnDBUpdate")
    self.db.RegisterCallback(self, "OnNewProfile", "OnDBUpdate")
    self.db.RegisterCallback(self, "OnProfileReset", "OnDBUpdate")
    self.db.RegisterCallback(self, "OnProfileChanged", "OnDBUpdate")    

    self:RegisterChatCommand("whisperinvite", "CMD")
    self:RegisterChatCommand("wi", "CMD")
    if L["CHAT_COMMAND"] ~= "CHAT_COMMAND" then
        self:RegisterChatCommand(L["CHAT_COMMAND"], "CMD")
    end

    if type(WhisperInviteSettings) == "table" and not WhisperInviteSettings.imported then
        for _, word in pairs(WhisperInviteSettings.keywords) do
            self.db.profile.keywords[word].active = true
            self.db.profile.keywords[word].fullMatch = WhisperInviteSettings.fullMatch
            self.db.profile.keywords[word].capitalization = WhisperInviteSettings.capitalization

            for channel, active in pairs(WhisperInviteSettings.channels) do
                self.db.profile.keywords[word].channels[channel] = active
            end           
        end

        local playerName = _G.UnitName("player"):utf8lower()
        for name, active in pairs(WhisperInviteSettings.blacklist.names) do
            if active and name:utf8lower() ~= playerName then
                self.db.global.blacklist.names[name] = true
            end
        end

        WhisperInviteSettings.imported = true
    end

    self:ReBuildLists()

    self:RegisterOptions()
end


-- build/rebuild white and blacklist
do
    local isWiped = {}
    function Addon:ReBuildLists()
        wipe(isWiped)
        for scope, active in pairs(ns.listScopes) do
            if active then
                for types, active in pairs(ns.listTypes) do
                    if active then
                        for entry, active in pairs(ns.listEntrys) do
                            if active then
                                if type(ns.lists[types][entry]) ~= "table" then
                                    ns.lists[types][entry] = {}
                                    isWiped[types..entry] = true
                                elseif not isWiped[types..entry] then
                                    isWiped[types..entry] = true
                                    wipe(ns.lists[types][entry])
                                end

                                if type(self.db[scope]) == "table" and type(self.db[scope][types]) == "table" and type(self.db[scope][types][entry]) == "table" then
                                    for key, active in pairs(self.db[scope][types][entry]) do
                                        if active then
                                            ns.lists[types][entry][key] = true
                                        end
                                    end
                                end
                            end
                        end
                    end
                end
            end
        end

        for id in pairs(ns.ID_cache) do
            ns.ID_cache[id]:Reset()
        end
    end    
end

function Addon:OnDBUpdate(event, db, profile)
    self:CleanOptions()
    
    self:ReBuildLists()

    self:Disable()
    if self.db.profile.active then
        self:Enable()
    end
end


function Addon:OnEnable()
    self:RegisterAllKeywords()
    --self:RegisterEvent("")
end

function Addon:OnDisable()
    self:UnregisterAllKeywords()
    --self:UnregisterEvent("")
end

function Addon:GetCacheEntry(type, key)
    return self.db.global.cache[type][key:utf8lower()] or key or "<Error: no key>"
end
function Addon:SetCacheEntry(type, key)
    self.db.global.cache[type][key:utf8lower()] = key
end

function Addon:ClearCache()
    for entry in pairs(self.db.global.cache) do
        for key in pairs(self.db.global.cache[entry]) do
            local found = false
            for scope, active in pairs(ns.listScopes) do
                if active then
                    for types, active in pairs(ns.listTypes) do
                        if active then
                            if type(self.db[scope]) == "table" and type(self.db[scope][types]) == "table" and type(self.db[scope][types][entry]) == "table" then
                                if nil ~= self.db[scope][types][entry][key] then
                                    found = true
                                    break
                                end
                            end
                        end
                    end
                end
                if found then
                    break
                end
            end

            if not found then
                self.db.global.cache[entry][key] = nil
            end            
        end
    end
end

do
    local list = {}

    function Addon:RegisterAllKeywords()
        for key in pairs(self.db.profile.keywords) do
            self:RegisterKeyword(key)
        end
    end
    function Addon:UnregisterAllKeywords()
        for event, key in pairs(list) do
            self:UnregisterEvent(event)
            list[event] = false
        end
    end
    function Addon:RegisterKeyword(keyword)
        if self.db.profile.keywords[keyword].active then
            for event, active in pairs(self.db.profile.keywords[keyword].channels) do
                if active then
                    if not list[event] then
                        list[event] = {}
                        self:RegisterEvent(event, "ChatMessageEvent")
                    end                    
                    list[event][keyword] = true
                end                
            end
        end
    end
    function Addon:UnregisterKeyword(keyword)
        for event in pairs(list) do
            if type(list[event]) == "table" then
                list[event][keyword] = nil
                if not next(list[event]) then
                    self:UnregisterEvent(event)
                end
            end
        end
    end

    function Addon:ChatMessageEvent(event, ...)        
        if _G.IsInGroup() and not _G.UnitIsGroupLeader("player") and not _G.UnitIsGroupAssistant("player") then return end

        local id
        local msg = select(1, ...)
        if event:sub(1,11) == "CHAT_MSG_BN" then
            id = select(13, ...)
        else            
            id = select(12, ...)
        end

        if ns.ID_cache[id] then end

        for keyword in pairs(list[event]) do
            if self:CheckMessage(msg, keyword) then
                if self:Invite(keyword, id) then
                    break
                end
            end
        end
    end

    function Addon:Invite(keyword, id, count)
        if ns.ID_cache[id].listsChecked then
            local inviteOK, reason, errFunc = ns.ID_cache[id]:CheckLists(self.db.profile.keywords[keyword].whitelist, self.db.profile.keywords[keyword].blacklist, true)
            if inviteOK and ns.ID_cache[id]:Invite(true) then
                if self.db.profile.keywords[keyword].autoRaid and _G.GetNumGroupMembers(_G.LE_PARTY_CATEGORY_HOME) == 4 then
                    _G.ConvertToRaid()
                end
                ns.ID_cache[id]:Invite()
                return true
            else
                if self.db.profile.keywords[keyword].showWhyNotInvited and type(errFunc) == "function" then
                    if reason then
                        ns.ID_cache[id][reason] = true
                        self:Print(errFunc(ns.ID_cache[id]) )
                        ns.ID_cache[id][reason] = nil
                    else
                        self:Print(errFunc(ns.ID_cache[id]) )
                    end                    
                end
                return false
            end
        else
            if ns.ID_cache[id]:Invite(true) then
                if count and count > self.db.profile.maxInviteScheudles then
                    self:Print(L["Abort invite schedule, reached maximal count for key: %q for %s."]:format(keyword, ns.ID_cache[id].name))
                else
                    if self.db.profile.keywords[keyword].showScheduleMessage and count and count > 1 then
                        self:Print(L["Schedule invite with key: %q for %s."]:format(keyword, ns.ID_cache[id].name))
                    end
                    self:ScheduleTimer("Invite", self.db.profile.keywords[keyword].scheduleTime, keyword, id, (count or 0)+1)
                end
                return false
            else
                if self.db.profile.keywords[keyword].showWhyNotInvited then
                    self:Print(ns.ID_cache[id]:Invite())
                end
                return false                
            end
        end
    end
end

function Addon:CheckMessage(msg, keyword)
    if not self.db.profile.keywords[keyword].capitalization then
        msg = msg:utf8lower()
        keyword = keyword:utf8lower()
    end
    if self.db.profile.keywords[keyword].fullMatch then
        if msg ~= keyword then
            return false
        end
    else
        if not msg:find(keyword) then
            return false
        end
    end    

    return true
end

-- LOCALES: CMD_HELP="Chat commands\n|cFF278BCB/wi|r - opens settings\n|cFF278BCB/wi|r clearcache - clears cache of unused black/whitelist entries\n|cFF278BCB/wi|r advance - set if should show advance settings" 
-- LOCALES: CMD_HELP_advance="|cFF278BCB/wi|r advance true or false"
-- LOCALES: CMD_clearcache="clearcache", CMD_advance="advance"
function Addon:CMD(input)
    local a1, lastPos = self:GetArgs(input, 1)
    if not a1 or a1 == "" then
        self:OpenSettings()
    elseif a1:lower() == "clearcache" or a1:utf8lower() == L["CMD_clearcache"]:utf8lower() then
        self:ClearCache()
    elseif a1:lower() == "advance" or a1:utf8lower() == L["CMD_advance"]:utf8lower() then
        local a2, lastPos = self:GetArgs(input, 1, lastPos)
        local number = tonumber(a2)
        if number then
            if number == 0 then
                self.profile.showAdvanceSettings = false
                self:Print(L["Advance settings are hidden."])
            else
                self.profile.showAdvanceSettings = true
                self:Print(L["Advance settings are shown."])
            end
        else
            if a2:lower() == "true" then
                self.profile.showAdvanceSettings = true
                self:Print(L["Advance settings are shown."])
            elseif a2:lower() == "false" then
                self.profile.showAdvanceSettings = false
                self:Print(L["Advance settings are hidden."])
            else
                self:Print(L["CMD_HELP_advance"])
            end
        end
    else        
        self:Print(L["CMD_HELP"])
    end
end

function Addon:OpenSettings()
    _G.InterfaceOptionsFrame_OpenToCategory(ADDONNAME)
    if not _G.InterfaceOptionsFrameAddOns:IsShown() then
        _G.InterfaceOptionsFrame_OpenToCategory(ADDONNAME)
    end
end

do
    local mainOptions = {}
    local globalList = {}
    local profileList = {}
    local keywordList = {}

    local listOptions = {}

    local dash = "-"

    local OptionsHandler = {}
    function OptionsHandler:Set(info, value, ...)
        info.arg[info[#info]] = value
    end
    function OptionsHandler:Get(info, value, ...)
        return info.arg[info[#info]]
    end
    function OptionsHandler:ValidateListInput(info, value, ...)
        if ns.listEntrys[info.arg.entryKey] then
            if value:trim():len() < 1 then
                return L["%s is a empty string."]:format(L[info.arg.entryKey])
            end
            if info.arg.entryKey == "names" then
                if dash:split(value):trim():find("%s") then
                    return L["Name can't include space characters."]
                end
            elseif info.arg.entryKey == "guilds" then
            elseif info.arg.entryKey == "realms" then
            end
            return true
        end
        return L["Can't validate. Entry type %s doesn't exits as List entry."]:format(L[info.arg.entryKey])
    end
    function OptionsHandler:NewListEntry(info, value, ... )
        local name = value:trim()        
        if info.arg.entryKey == "names" then
            local realm, realm2
            name, realm, realm2 = dash:split(name)        
            if realm then                
                name = name:trim().."-"..realm:trim()

                if realm2 then
                    name = name:trim().."-"..realm2:trim()
                end
            end
        end

        Addon:SetCacheEntry(info.arg.entryKey, name)

        info.arg.db[info.arg.typeKey][info.arg.entryKey][name:utf8lower()] = true
        Addon:ReBuildLists()
    end
    function OptionsHandler:GetEmptyString(info, value, ... )        
        return ""
    end
    function OptionsHandler:SetListEntry(...)
        self:Set(...)
        Addon:ReBuildLists()
    end
    function OptionsHandler:DeleteInactiveEntries(info, value, ...)
        if info.arg and info.arg.db then
            for typeKey in pairs(ns.listTypes) do
                for entryKey in pairs(ns.listEntrys) do

                    for key, value in pairs(info.arg.db[typeKey][entryKey]) do
                        if not value then
                            info.arg.db[typeKey][entryKey][key] = nil

                            for id in pairs(listOptions) do 
                                listOptions[id][typeKey].args[entryKey].args[key] = nil
                            end
                        end
                    end
                end
            end
        end        
    end
    function OptionsHandler:ValidateKeywordInput(info, value, ...)
        if value and value:trim():len() > 0 then            
            return true
        end
        return L["Keyword is not valid."]
    end
    function OptionsHandler:NewKeywordEntry(info, value, ...)        
        info.arg[value:trim()].exists = true
    end
    function OptionsHandler:GetSelect(info, ...)
        return info.arg[select(select('#', ...), ...)]
    end
    function OptionsHandler:SetSelect(info, key, value, ...)
        info.arg[key] = value
    end
    local SelectListCache = {}
    function OptionsHandler:SelectValues(info, value, ...)
        SelectListCache[info.arg] = SelectListCache[info.arg] or {}
        for key in pairs(info.arg) do
            SelectListCache[info.arg][key] = L[key]
        end
        return SelectListCache[info.arg]
    end
    function OptionsHandler:DeleteKeyword(info, value, ...)
        Addon.db.profile.keywords[info.arg] = nil
        mainOptions.args[info.arg] = nil        
    end
    function OptionsHandler:ValidateChannelInput(info, value, ...)
        if value and value:trim():len() > 0 then
            if value:sub(1,9):upper() ~= "CHAT_MSG_" then
                return L["Invalid channel event name."]
            end
            return true
        end
        return L["Channel Event is not valid."]
    end
    function OptionsHandler:NewChannelEntry(info, value, ...)
        info.arg[value:upper()] = true
    end


    function Addon.GetListsOptions(uiType, uiName, appName, order, ...)
        local id = ""
        local dbCount = select('#', ...)
        for i=1, dbCount do
            id = id..tostring(select(i, ...))
        end

        local options = listOptions[id] or {}

        if type(order) ~= "number" then
            order = 100
        end

        local i = 0.1
        options.listDescription = options.listDescription or {}
        options.listDescription.name = L["When a Whitelist entry exists only people in the list will be invited.\nBlacklist filtering takes affect after Whitelist filtering, e.g: When a player has a Whitelist entry in Name but the Guild is blacklisted s/he will not be invited."]
        options.listDescription.type = "description"
        options.listDescription.fontSize = "small"
        options.listDescription.width = "full"
        options.listDescription.order = order + i
        i = i + 0.1

        options.removeInactiveEntries = options.removeInactiveEntries or {}
        options.removeInactiveEntries.name = L["Delete Inactive Entries"]
        options.removeInactiveEntries.desc = L["Delete all Inactive Entries"]
        options.removeInactiveEntries.type = "execute"
        options.removeInactiveEntries.func = "DeleteInactiveEntries"
        options.removeInactiveEntries.arg = options.removeInactiveEntries.arg or {}
        options.removeInactiveEntries.arg.db = select(1, ...)
        options.removeInactiveEntries.arg.id = id
        options.removeInactiveEntries.order = order + i
        i = i + 0.1

        for typeKey in pairs(ns.listTypes) do
            options[typeKey] = options[typeKey] or {}
            options[typeKey].name = L[typeKey]
            options[typeKey].type = "group"
            options[typeKey].childGroups = "tab"
            options[typeKey].set = "SetListEntry"
            options[typeKey].order = order + i
            options[typeKey].args = options[typeKey].args or {}

            local j = 1
            for entryKey in pairs(ns.listEntrys) do
                options[typeKey].args[entryKey] = options[typeKey].args[entryKey] or {}
                options[typeKey].args[entryKey].name = L[entryKey]
                options[typeKey].args[entryKey].type = "group"
                options[typeKey].args[entryKey].childGroups = "tree"
                options[typeKey].args[entryKey].order = 1
                options[typeKey].args[entryKey].args = options[typeKey].args[entryKey].args or {}

                options[typeKey].args[entryKey].args.inputNewEntry = options[typeKey].args[entryKey].args.inputNewEntry or {}
                options[typeKey].args[entryKey].args.inputNewEntry.name = L["Add %s"]:format(entryKey)
                options[typeKey].args[entryKey].args.inputNewEntry.desc = options[typeKey].args[entryKey].args.inputNewEntry.desc or L["Add a new %s to the %s"]:format(L[entryKey], L[typeKey])
                options[typeKey].args[entryKey].args.inputNewEntry.type = "input"
                options[typeKey].args[entryKey].args.inputNewEntry.validate = "ValidateListInput"
                options[typeKey].args[entryKey].args.inputNewEntry.arg = options[typeKey].args[entryKey].args.inputNewEntry.arg or {}
                options[typeKey].args[entryKey].args.inputNewEntry.arg.typeKey = typeKey
                options[typeKey].args[entryKey].args.inputNewEntry.arg.entryKey = entryKey
                options[typeKey].args[entryKey].args.inputNewEntry.arg.db = select(1,...)
                options[typeKey].args[entryKey].args.inputNewEntry.set = "NewListEntry"
                options[typeKey].args[entryKey].args.inputNewEntry.get = "GetEmptyString"
                options[typeKey].args[entryKey].args.inputNewEntry.width = "full"
                options[typeKey].args[entryKey].args.inputNewEntry.order = 1

                local k = 2
                if dbCount > 1 then
                    options[typeKey].args[entryKey].args.inputDescription = options[typeKey].args[entryKey].args.inputDescription or {}
                    options[typeKey].args[entryKey].args.inputDescription.name = L["Entries in |cFFFFD100this colour|r are overwritten by lists with higher priorities.\nEntries in |cFF888888this colour|r exists in list with higher priorities."]
                    options[typeKey].args[entryKey].args.inputDescription.type = "description"
                    options[typeKey].args[entryKey].args.inputDescription.fontSize = "small"
                    options[typeKey].args[entryKey].args.inputDescription.width = "full"
                    options[typeKey].args[entryKey].args.inputDescription.order = k
                    k = k + 1
                end
                
                for l=1, dbCount do
                    local db = select(l, ...)
                    if type(db) == "table" then
                        for key,value in pairs(db[typeKey][entryKey]) do
                            options[typeKey].args[entryKey].args[key] = options[typeKey].args[entryKey].args[key] or {}
                            
                            options[typeKey].args[entryKey].args[key].name = Addon:GetCacheEntry(entryKey, key)
                            options[typeKey].args[entryKey].args[key].desc = options[typeKey].args[entryKey].args[key].desc or L["Is this %s in the %s."]:format(L[entryKey], L[typeKey])
                            options[typeKey].args[entryKey].args[key].type = "toggle"

                            if l == 1 then                                
                                options[typeKey].args[entryKey].args[key].arg = db[typeKey][entryKey]
                                options[typeKey].args[entryKey].args[key].disabled = false
                            elseif options[typeKey].args[entryKey].args[key].arg == db[typeKey][entryKey] or not options[typeKey].args[entryKey].args[key].arg then                                
                                options[typeKey].args[entryKey].args[key].disabled = true
                                options[typeKey].args[entryKey].args[key].arg = db[typeKey][entryKey]
                            end

                            if l > 1 and not options[typeKey].args[entryKey].args[key].disabled then
                                options[typeKey].args[entryKey].args[key].name = "|cFFFFD100"..Addon:GetCacheEntry(entryKey, key).."|r"
                            end
                            
                            if not options[typeKey].args[entryKey].args[key].order then
                                options[typeKey].args[entryKey].args[key].order = k
                                k = k + 1
                            end                            
                        end
                    end
                end                

                j = j + 1
            end

            i = i + 0.1
        end
        
        listOptions[id] = options
        return listOptions[id]
    end

    function Addon.GetMainOptionsTable(uiType, uiName, appName)
        mainOptions.name = L["Keywords"]
        mainOptions.type = "group"
        mainOptions.childGroups = "select"
        mainOptions.desc = L["WhisperInvite main settings."]
        mainOptions.handler = OptionsHandler
        mainOptions.set = "Set"
        mainOptions.get = "Get"
        mainOptions.args = mainOptions.args or {}

        mainOptions.args.active = mainOptions.args.active or {}
        mainOptions.args.active.name = L["Active"]
        mainOptions.args.active.type = "toggle"
        mainOptions.args.active.desc = L["If this Profile active."]
        mainOptions.args.active.arg = Addon.db.profile        
        mainOptions.args.active.width = "full"
        mainOptions.args.active.order = 1

        mainOptions.args.inputNewKeyword = mainOptions.args.inputNewKeyword or {}
        mainOptions.args.inputNewKeyword.name = L["Add Keyword"]
        mainOptions.args.inputNewKeyword.desc = L["Create a new Keyword"]
        mainOptions.args.inputNewKeyword.type = "input"
        mainOptions.args.inputNewKeyword.validate = "ValidateKeywordInput"
        mainOptions.args.inputNewKeyword.arg = Addon.db.profile.keywords
        mainOptions.args.inputNewKeyword.set = "NewKeywordEntry"
        mainOptions.args.inputNewKeyword.get = "GetEmptyString"
        mainOptions.args.inputNewKeyword.order = 2        

        local i = 3
        for keyword, value in pairs(Addon.db.profile.keywords) do
            local db = Addon.db.profile.keywords[keyword]

            mainOptions.args[keyword] = mainOptions.args[keyword] or {}
            mainOptions.args[keyword].name = keyword
            mainOptions.args[keyword].type = "group"
            mainOptions.args[keyword].childGroups = "tab"
            mainOptions.args[keyword].order = i

            mainOptions.args[keyword].args = mainOptions.args[keyword].args or {}

            mainOptions.args[keyword].args.main = mainOptions.args[keyword].args.main or {}
            mainOptions.args[keyword].args.main.name = L["Settings"]
            mainOptions.args[keyword].args.main.type = "group"
            mainOptions.args[keyword].args.main.order = 1
            mainOptions.args[keyword].args.main.args = mainOptions.args[keyword].args.main.args or {}

            mainOptions.args[keyword].args.main.args.active = mainOptions.args[keyword].args.main.args.active or {}
            mainOptions.args[keyword].args.main.args.active.name = L["Active"]
            mainOptions.args[keyword].args.main.args.active.desc = L["If this keyword active."]
            mainOptions.args[keyword].args.main.args.active.type = "toggle"
            mainOptions.args[keyword].args.main.args.active.width = "full"
            mainOptions.args[keyword].args.main.args.active.arg = db
            mainOptions.args[keyword].args.main.args.active.order = 1

            mainOptions.args[keyword].args.main.args.fullMatch = mainOptions.args[keyword].args.main.args.fullMatch or {}
            mainOptions.args[keyword].args.main.args.fullMatch.name = L["Exact match"]
            mainOptions.args[keyword].args.main.args.fullMatch.desc = L["The keyword must match exact.\nWhen not, the keyword will be searched in the entire message.\nNote: When not exact match the keyword will be interpreter as pattern."]
            mainOptions.args[keyword].args.main.args.fullMatch.type = "toggle"
            mainOptions.args[keyword].args.main.args.fullMatch.arg = db
            mainOptions.args[keyword].args.main.args.fullMatch.width = "double"
            mainOptions.args[keyword].args.main.args.fullMatch.order = 2

            mainOptions.args[keyword].args.main.args.capitalization = mainOptions.args[keyword].args.main.args.capitalization or {}
            mainOptions.args[keyword].args.main.args.capitalization.name = L["Case sensitive"]            
            mainOptions.args[keyword].args.main.args.capitalization.type = "toggle"
            mainOptions.args[keyword].args.main.args.capitalization.arg = db
            mainOptions.args[keyword].args.main.args.capitalization.width = "double"
            mainOptions.args[keyword].args.main.args.capitalization.order = 3

            mainOptions.args[keyword].args.main.args.showWhyNotInvited = mainOptions.args[keyword].args.main.args.showWhyNotInvited or {}
            mainOptions.args[keyword].args.main.args.showWhyNotInvited.name = L["Not invited message"]
            mainOptions.args[keyword].args.main.args.showWhyNotInvited.desc = L["Show why not invited message."]
            mainOptions.args[keyword].args.main.args.showWhyNotInvited.type = "toggle"
            mainOptions.args[keyword].args.main.args.showWhyNotInvited.arg = db
            mainOptions.args[keyword].args.main.args.showWhyNotInvited.width = "double"
            mainOptions.args[keyword].args.main.args.showWhyNotInvited.order = 4

            mainOptions.args[keyword].args.main.args.autoRaid = mainOptions.args[keyword].args.main.args.autoRaid or {}
            mainOptions.args[keyword].args.main.args.autoRaid.name = L["Auto Convert"]
            mainOptions.args[keyword].args.main.args.autoRaid.desc = L["Convert automatic to raid group if the sixth player will be invited."]
            mainOptions.args[keyword].args.main.args.autoRaid.type = "toggle"
            mainOptions.args[keyword].args.main.args.autoRaid.arg = db
            mainOptions.args[keyword].args.main.args.autoRaid.width = "double"
            mainOptions.args[keyword].args.main.args.autoRaid.order = 4.5

            mainOptions.args[keyword].args.main.args.showScheduleMessage = mainOptions.args[keyword].args.main.args.showScheduleMessage or {}
            mainOptions.args[keyword].args.main.args.showScheduleMessage.name = L["Schedule message"]
            mainOptions.args[keyword].args.main.args.showScheduleMessage.desc = L["Show message when a invite is scheduled."]
            mainOptions.args[keyword].args.main.args.showScheduleMessage.type = "toggle"
            mainOptions.args[keyword].args.main.args.showScheduleMessage.arg = db
            mainOptions.args[keyword].args.main.args.showScheduleMessage.width = "double"
            mainOptions.args[keyword].args.main.args.showScheduleMessage.order = 5

            mainOptions.args[keyword].args.main.args.scheduleTime = mainOptions.args[keyword].args.main.args.scheduleTime or {}
            mainOptions.args[keyword].args.main.args.scheduleTime.name = L["Invite schedule time"]
            mainOptions.args[keyword].args.main.args.scheduleTime.desc = L["Time between invite check schedule."]
            mainOptions.args[keyword].args.main.args.scheduleTime.type = "range"
            mainOptions.args[keyword].args.main.args.scheduleTime.min = 0.5
            mainOptions.args[keyword].args.main.args.scheduleTime.max = 10
            mainOptions.args[keyword].args.main.args.scheduleTime.softMin = 1
            mainOptions.args[keyword].args.main.args.scheduleTime.softMax = 5
            mainOptions.args[keyword].args.main.args.scheduleTime.step = 0.1
            mainOptions.args[keyword].args.main.args.scheduleTime.bigStep = 0.5            
            mainOptions.args[keyword].args.main.args.scheduleTime.arg = db
            mainOptions.args[keyword].args.main.args.scheduleTime.order = 6

            mainOptions.args[keyword].args.main.args.deleteKeyword = mainOptions.args[keyword].args.main.args.deleteKeyword or {}
            mainOptions.args[keyword].args.main.args.deleteKeyword.name = L["Delete Keyword"]
            mainOptions.args[keyword].args.main.args.deleteKeyword.desc = L["Delete this Keyword."]
            mainOptions.args[keyword].args.main.args.deleteKeyword.type = "execute"
            mainOptions.args[keyword].args.main.args.deleteKeyword.func = "DeleteKeyword"
            mainOptions.args[keyword].args.main.args.deleteKeyword.arg = keyword
            mainOptions.args[keyword].args.main.args.deleteKeyword.order = 7


            mainOptions.args[keyword].args.channels = mainOptions.args[keyword].args.channels or {}
            mainOptions.args[keyword].args.channels.name = L["Channels"]
            mainOptions.args[keyword].args.channels.type = "group"
            mainOptions.args[keyword].args.channels.order = 2
            mainOptions.args[keyword].args.channels.args = mainOptions.args[keyword].args.channels.args or {}

            mainOptions.args[keyword].args.channels.args.inputNewChannel = mainOptions.args[keyword].args.channels.args.inputNewChannel or {}
            mainOptions.args[keyword].args.channels.args.inputNewChannel.name = L["Add Channel"]
            mainOptions.args[keyword].args.channels.args.inputNewChannel.desc = L["Add new channel event for listening.\nChannel event must begin with %q.\n"]:format("CHAT_MSG_")
            mainOptions.args[keyword].args.channels.args.inputNewChannel.type = "input"            
            mainOptions.args[keyword].args.channels.args.inputNewChannel.validate = "ValidateChannelInput"
            mainOptions.args[keyword].args.channels.args.inputNewChannel.set = "NewChannelEntry"
            mainOptions.args[keyword].args.channels.args.inputNewChannel.get = "GetEmptyString"
            mainOptions.args[keyword].args.channels.args.inputNewChannel.arg = db.channels
            mainOptions.args[keyword].args.channels.args.inputNewChannel.hidden = not Addon.db.profile.showAdvanceSettings
            mainOptions.args[keyword].args.channels.args.inputNewChannel.order = 1

            mainOptions.args[keyword].args.channels.args.channels = mainOptions.args[keyword].args.channels.args.channels or {}
            mainOptions.args[keyword].args.channels.args.channels.name = L["Channels"]
            mainOptions.args[keyword].args.channels.args.channels.desc = L["Channels where this keyword can be used."]
            mainOptions.args[keyword].args.channels.args.channels.type = "multiselect"
            mainOptions.args[keyword].args.channels.args.channels.style = "dropdown"
            mainOptions.args[keyword].args.channels.args.channels.values = "SelectValues"
            mainOptions.args[keyword].args.channels.args.channels.get = "GetSelect"
            mainOptions.args[keyword].args.channels.args.channels.set = "SetSelect"
            mainOptions.args[keyword].args.channels.args.channels.arg = db.channels
            mainOptions.args[keyword].args.channels.args.channels.order = 2


            mainOptions.args[keyword].args.lists = mainOptions.args[keyword].args.lists or {}
            mainOptions.args[keyword].args.lists.name = L["Lists"]
            mainOptions.args[keyword].args.lists.type = "group"
            mainOptions.args[keyword].args.lists.childGroups = "tab"
            mainOptions.args[keyword].args.lists.order = 3
            mainOptions.args[keyword].args.lists.args = mainOptions.args[keyword].args.lists.args or {}         
            mainOptions.args[keyword].args.lists.plugins = mainOptions.args[keyword].args.lists.plugins or {}
            mainOptions.args[keyword].args.lists.plugins["Lists"] = Addon.GetListsOptions(uiType, uiName, appName, 1, db, Addon.db.profile, Addon.db.global)

            i = i + 1
        end

        return mainOptions
    end

    function Addon.GetGlobalListsOptionsTable(uiType, uiName, appName)
        globalList.name = L["Global Lists"]
        globalList.type = "group"
        globalList.childGroups = "tab"
        globalList.desc = L["Global Blacklist and Whitelist settings."]
        globalList.handler = OptionsHandler
        globalList.set = "Set"
        globalList.get = "Get"        
        globalList.args = globalList.args or {}

        globalList.plugins = globalList.plugins or {}
        globalList.plugins["Lists"] = Addon.GetListsOptions(uiType, uiName, appName, 1, Addon.db.global)

        return globalList
    end
    
    function Addon.GetProfileListOptionsTable(uiType, uiName, appName)
        profileList.name = L["Profile Lists"]
        profileList.type = "group"
        profileList.childGroups = "tab"
        profileList.desc = L["Profile Blacklist and Whitelist settings."]
        profileList.handler = OptionsHandler
        profileList.set = "Set"
        profileList.get = "Get"        
        profileList.args = profileList.args or {}

        profileList.plugins = profileList.plugins or {}
        profileList.plugins["Lists"] = Addon.GetListsOptions(uiType, uiName, appName, 1, Addon.db.profile, Addon.db.global)

        return profileList
    end

    function Addon:RegisterOptions()
        AC:RegisterOptionsTable(ADDONNAME, self.GetMainOptionsTable)
        AC:RegisterOptionsTable(ADDONNAME.."_GlobalLists", self.GetGlobalListsOptionsTable)
        AC:RegisterOptionsTable(ADDONNAME.."_ProfileLists", self.GetProfileListOptionsTable)

        AC:RegisterOptionsTable(ADDONNAME.."_Profile", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db) )
        
        -- LOCALE: WhisperInvite
        ACD:AddToBlizOptions(ADDONNAME, L[ADDONNAME])
        ACD:AddToBlizOptions(ADDONNAME.."_GlobalLists", L["Global Lists"], L[ADDONNAME])
        ACD:AddToBlizOptions(ADDONNAME.."_ProfileLists", L["Profile Lists"], L[ADDONNAME])

        ACD:AddToBlizOptions(ADDONNAME.."_Profile", L["Profile"], L[ADDONNAME])
    end

    function Addon:CleanOptions()
        wipe(profileList)
        wipe(mainOptions)
    end
end



